import os.path

from face2face import Face2Face
import cv2
from media_toolkit import VideoFile, ImageFile


f2f = Face2Face(device_id=0)

output_folder = "output"

def test_single_face_swap():
    source_img = "test_media/test_face_1.jpg"
    target_img = "test_media/test_face_2.jpg"
    swapped = f2f.swap_img_to_img(source_img, target_img, enhance_face_model=None)
    ImageFile().from_np_array(swapped).save(f"{output_folder}/test_swap.jpg")


def test_multi_face_swap():
    source_img = "test_media/three.jpeg"
    target_img = "test_media/three_men.jpeg"
    swapped = f2f.swap_img_to_img(source_img, target_img, enhance_face_model='gpen_bfr_2048')
    cv2.imwrite(f"{output_folder}/multi_swap.png", swapped)

def test_embedding_face_swap():
    source_img = "test_media/test_face_1.jpg"
    target_img = "test_media/test_face_2.jpg"
    f2f.add_face("hagrid", source_img, save=True)
    swapped = f2f.swap(media=target_img, faces="hagrid")
    cv2.imwrite(f"{output_folder}/test_swap.png", swapped)

def test_multi_face_from_reference():
    f1 = "test_media/test_face_1.jpg"
    f2 = "test_media/test_face_2.jpg"
    f3 = "test_media/test_face_3.png"
    f4 = "test_media/test_face_4.jpg"
    f2f.add_face("hagrid", f1, save=True)
    f2f.add_face("ron", f2, save=True)
    f2f.add_face("biden", f3, save=True)
    f2f.add_face("trump", f4, save=True)

    test_multi_face_img = "test_media/test_multi_swap_from_reference.jpg"
    #swapped = f2f.swap_pairs(test_multi_face_img, swap_pairs={
    #    "trump": "hagrid",
    #    "biden": "ron"
    #})
    swapped = f2f.swap(media=test_multi_face_img, faces={
        "trump": "hagrid",
        "biden": "ron"
    })
    cv2.imwrite("../docs/swap_with_recognition.jpg", swapped)


def test_face_enhancing():
    source_img = "test_media/test_face_1.jpg"
    enhanced = f2f.enhance_faces(image=source_img, model='gpen_bfr_2048')
    cv2.imwrite(f"{output_folder}/enhance_test_gpen_bfr_2048.png", enhanced)

def test_face_enhancing_single_face():
    source_img = "test_media/three.jpeg"
    f = f2f.detect_faces(source_img)
    enhanced = f2f.enhance_single_face(image=source_img, target_face=f[0], model='gfpgan_1.4')
    cv2.imwrite(f"{output_folder}/enhance_single_face_gfpgan_1.4.png", enhanced)

def test_video_face_swap():
    # add ref face
    source_img = cv2.imread("test_media/test_face_4.jpg")
    # f2f.add_face("caprio", source_img, save=True)
    # swap it
    fn = "test_media/test_video_ultra_short_short.mp4"
    vf = VideoFile().from_file(fn)
    #swapped = f2f.swap_to_face_in_video(face_name="caprio", video=vf)
    swapped = f2f.swap(media=vf, faces="caprio", enhance_face_model=None)
    swapped.save(f"{output_folder}/{os.path.basename(fn)}_swapped.mp4")
    a = 1

def test_multi_face_video_swap():
    # add ref face
    f2f.add_face("biden", "test_media/test_face_3.png", save=True)
    f2f.add_face("harris", "test_media/test_face_5.jpg", save=True)
    f2f.add_face("murphy", "test_media/test_face_6.jpg", save=True)
    f2f.add_face("warren", "test_media/test_face_7.jpeg", save=True)
    # swap it
    vf = VideoFile().from_file("test_media/test_video_1.mp4")
    #swapped = f2f.swap_pairs_in_video(swap_pairs={
    #    "murphy": "biden",
    #    "warren": "harris"
    #}, video=vf)
    swapped = f2f.swap(media=vf, enhance_face_model=None, faces={
        "murphy": "biden",
        "warren": "harris",
    })
    swapped.save(f"{output_folder}/test_media/z.mp4")


if __name__ == "__main__":
    test_single_face_swap()
    # test_multi_face_swap()
    #test_multi_face_from_reference()
    #test_face_enhancing()
    #test_face_enhancing_single_face()
    #test_multi_face_video_swap()
    # test_embedding_face_swap()
    test_video_face_swap()
    a = 1




def create_occlusion_mask(crop_vision_frame: VisionFrame) -> Mask:
    face_occluder = get_face_occluder()
    prepare_vision_frame = cv2.resize(crop_vision_frame, face_occluder.get_inputs()[0].shape[1:3][::-1])
    prepare_vision_frame = numpy.expand_dims(prepare_vision_frame, axis=0).astype(numpy.float32) / 255
    prepare_vision_frame = prepare_vision_frame.transpose(0, 1, 2, 3)
    with conditional_thread_semaphore(facefusion.globals.execution_providers):
        occlusion_mask: Mask = face_occluder.run(None,
                                                 {
                                                     face_occluder.get_inputs()[0].name: prepare_vision_frame
                                                 })[0][0]
    occlusion_mask = occlusion_mask.transpose(0, 1, 2).clip(0, 1).astype(numpy.float32)
    occlusion_mask = cv2.resize(occlusion_mask, crop_vision_frame.shape[:2][::-1])
    occlusion_mask = (cv2.GaussianBlur(occlusion_mask.clip(0, 1), (0, 0), 5).clip(0.5, 1) - 0.5) * 2
    return occlusion_mask



# avoid circular dependency but provide type hints
from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Tuple

from face2face.core.compatibility.Face import Face
from media_toolkit import ImageFile

if TYPE_CHECKING:
    from face2face.core.face2face import Face2Face

# regular imports
import glob
import os

import numpy as np


from face2face.core.modules.storage.f2f_loader import load_reference_face_from_file
from face2face.core.modules.storage.file_writable_face import FileWriteableFace
from face2face.settings import EMBEDDINGS_DIR
from face2face.core.modules.utils.utils import encode_path_safe
from face2face.core.modules.utils.utils import load_image


class _FaceEmbedding:
    def load_face(self: Face2Face, face_name: str) -> Face:
        """
        Load a reference face embedding from a file.
        :param face_name: the name of the reference face embedding
        :return: the embedding of the reference face(s)
        """
        # check if is already in ram. If yes return that one
        embedding = self._face_embeddings.get(face_name, None)
        if embedding is not None:
            return embedding

        # load from file
        file = os.path.join(self._face_embedding_folder, f"{face_name}.npy")
        face = load_reference_face_from_file(file)

        # add to memory dict
        self._face_embeddings[face_name] = face
        return face

    def load_faces(self, face_names: Union[str, List[str], List[Face], None] = None) -> dict:
        """
        :param face_names: the faces to load from the _face_embeddings folder.
            If None all stored face_embeddings are loaded and returned.
            If list of strings, the faces with the names in the list are loaded.
            If list of Face objects, the faces are returned as { index: face }.
        :return: the loaded faces as dict {face_name: face_embedding}.
        """
        if face_names is None:
            return self.load_all_faces()
        elif isinstance(face_names, str):
            return {face_names: self.load_face(face_names)}

        # convert whatever list to dict
        ret = {}
        for i, face in enumerate(face_names):
            if isinstance(face, Face):
                ret[i] = face
            elif isinstance(face, str):
                ret[face] = self.load_face(face)

        return ret

    def load_all_faces(self: Face2Face):
        """
        Load all face embeddings from the _face_embeddings folder.
        """
        for face_name in glob.glob(self._face_embedding_folder + "/*.npy"):
            self.load_face(face_name)
        return self._face_embeddings

    def add_face(
        self: Face2Face,
        face_name: Union[str, List[str]],
        image: Union[np.array, str, ImageFile],
        save: bool = False
    ) -> Union[Tuple[str, FileWriteableFace], dict]:
        """
        Add one or multiple reference face(s) to the face swapper. This face(s) can be used for swapping in other images.

        :param face_name: The name for the reference face.
            In case you provide a list of face names, an embedding is stored for each face from left to right in the provided image.

        :param image: The image from which to extract the face(s) (can be a numpy array, file path, or ImageFile).
            If there are multiple faces in the image, an embedding will be created for each name from left to right.
            If you only provide one name, only the first face will be stored.
        :param save:
            If True, the reference face will be saved to the _face_embeddings folder for future use.
            If False, the reference face will only be stored in memory.
        :return: A tuple containing the safely encoded face name and the reference face.
            In case of multiple faces returns as dict {face_name: face_embedding}
        """


        if isinstance(face_name, list) and len(face_name) == 0 or face_name is None:
            raise ValueError("Please provide at least one face name.")

        try:
            image = load_image(image)
            detected_faces = self.detect_faces(image)
        except Exception as e:
            raise ValueError(f"Could not load image or face detection failed. Error: {e}")

        # Deal with errors of too much or too few faces in the reference image
        if not detected_faces:
            raise ValueError(f"No faces detected in the provided image for {face_name}.")

        if not isinstance(face_name, list):
            face_name = [face_name]

        if len(face_name) > len(detected_faces):
            print(f"Not enough faces in the image for all provided face names. "
                  f"Only {len(detected_faces)} faces found. Skipping the rest")
            face_name = face_name[:len(detected_faces)]

        face = None
        for i, name in enumerate(face_name):
            name = encode_path_safe(name)
            face = detected_faces[i]
            # Store the detected faces in memory
            self._face_embeddings[name] = face

            # store the detected faces on disc
            face = FileWriteableFace(face)
            # Save face to virtual file
            if save:
                os.makedirs(EMBEDDINGS_DIR, exist_ok=True)
                filename = os.path.join(EMBEDDINGS_DIR, f"{name}.npy")
                if os.path.isfile(filename):
                    print(f"Reference face {name} already exists. Overwriting.")
                face.to_file(filename)
        if len(face_name) > 1:
            # Return a dict with the face names and embeddings
            return {
                name: self._face_embeddings[name]
                for name in face_name
            }
        else:
            return face_name[0], face



import os

from tqdm import tqdm

from face2face.settings import MODELS_DIR
from face2face.core.modules.utils import get_files_in_dir



def upscale_images_in_folder(image_folder: str, outpath: str = None):
    print(f"upscaling images in {image_folder}")
    path_to_real_esrgan = (
        MODELS_DIR + "/upscaling/realesrgan-ncnn-vulkan/realesrgan-ncnn-vulkan.exe"
    )

    # get imags
    lowres_imgs = get_files_in_dir(image_folder, [".jpeg", ".jpg", ".png"])
    # order images by creation date otherwise the video will be out of order
    lowres_imgs.sort(key=os.path.getmtime)

    # create output dir
    if outpath is None:
        outpath = os.path.join(image_folder, "upscaled")
        if not os.path.isdir(outpath):
            os.makedirs(outpath)

    # convert images with realesrgan
    for limg in tqdm(lowres_imgs):
        out_img = outpath + "/" + os.path.basename(limg)
        cmd = f"{path_to_real_esrgan} -i {limg} -o {out_img}"
        os.system(cmd)


def upscale_video(video_path: str, outpath: str = None):
    """
    Uses ESRGAN to upscale video. The audio is reaplied to the upscaled video.
    """
    # video2images(video_path, outpath)
    # extract_audio_from_video(video_path, outpath)
    # upscale images
    outupscaled = outpath + "/upscaled"
    # upscale_images_in_folder(outpath, outupscaled)

    # make video from images
    image_paths = get_files_in_dir(outupscaled, [".jpeg", ".jpg", ".png"])
    # get framerate from file if it exists
    if os.path.isfile(os.path.join(outpath, "_framerate.txt")):
        with open(os.path.join(outpath, "_framerate.txt"), "r") as f:
            frame_rate = int(float(f.readline()))
    else:
        print("Warning: Could not find framerate file. Using default framerate 60.")
        frame_rate = 60

    make_video_from_images(
        image_paths, outupscaled + "/upscaled.mp4", frame_rate=frame_rate
    )



# -*- coding: utf-8 -*-
from __future__ import division
import numpy as np
import cv2
import onnx
import onnxruntime
from face2face.core.compatibility import transform


class Attribute:
    def __init__(self, model_file=None, providers=None):
        assert model_file is not None
        self.model_file = model_file
        find_sub = False
        find_mul = False
        model = onnx.load(self.model_file)
        graph = model.graph
        for nid, node in enumerate(graph.node[:8]):
            #print(nid, node.name)
            if node.name.startswith('Sub') or node.name.startswith('_minus'):
                find_sub = True
            if node.name.startswith('Mul') or node.name.startswith('_mul'):
                find_mul = True
            if nid<3 and node.name=='bn_data':
                find_sub = True
                find_mul = True
        if find_sub and find_mul:
            #mxnet arcface model
            input_mean = 0.0
            input_std = 1.0
        else:
            input_mean = 127.5
            input_std = 128.0
        self.input_mean = input_mean
        self.input_std = input_std

        self.session = onnxruntime.InferenceSession(self.model_file, providers=providers)
        input_cfg = self.session.get_inputs()[0]
        input_shape = input_cfg.shape
        input_name = input_cfg.name
        self.input_size = tuple(input_shape[2:4][::-1])
        self.input_shape = input_shape
        outputs = self.session.get_outputs()
        output_names = []
        for out in outputs:
            output_names.append(out.name)
        self.input_name = input_name
        self.output_names = output_names
        assert len(self.output_names)==1
        output_shape = outputs[0].shape
        #print('init output_shape:', output_shape)
        if output_shape[1]==3:
            self.taskname = 'genderage'
        else:
            self.taskname = 'attribute_%d'%output_shape[1]

    def prepare(self, ctx_id, **kwargs):
        if ctx_id<0:
            self.session.set_providers(['CPUExecutionProvider'])

    def get(self, img, face):
        bbox = face.bbox
        w, h = (bbox[2] - bbox[0]), (bbox[3] - bbox[1])
        center = (bbox[2] + bbox[0]) / 2, (bbox[3] + bbox[1]) / 2
        rotate = 0
        _scale = self.input_size[0]  / (max(w, h)*1.5)
        #print('param:', img.shape, bbox, center, self.input_size, _scale, rotate)
        aimg, M = transform.transform(img, center, self.input_size[0], _scale, rotate)
        input_size = tuple(aimg.shape[0:2][::-1])
        #assert input_size==self.input_size
        blob = cv2.dnn.blobFromImage(aimg, 1.0/self.input_std, input_size, (self.input_mean, self.input_mean, self.input_mean), swapRB=True)
        pred = self.session.run(self.output_names, {self.input_name : blob})[0][0]
        if self.taskname=='genderage':
            assert len(pred)==3
            gender = np.argmax(pred[:2])
            age = int(np.round(pred[2]*100))
            face['gender'] = gender
            face['age'] = age
            return gender, age
        else:
            return pred



# -*- coding: utf-8 -*-
# @Organization  : insightface.ai
# @Author        : Jia Guo
# @Time          : 2021-05-04
# @Function      :

from __future__ import division
import numpy as np
import cv2
import onnx
import onnxruntime
from face2face.core.compatibility import face_align


class ArcFaceONNX:
    def __init__(self, model_file=None, providers=None):
        assert model_file is not None
        self.model_file = model_file
        self.taskname = 'recognition'
        find_sub = False
        find_mul = False
        model = onnx.load(self.model_file)
        graph = model.graph
        for nid, node in enumerate(graph.node[:8]):
            # print(nid, node.name)
            if node.name.startswith('Sub') or node.name.startswith('_minus'):
                find_sub = True
            if node.name.startswith('Mul') or node.name.startswith('_mul'):
                find_mul = True
        if find_sub and find_mul:
            # mxnet arcface model
            input_mean = 0.0
            input_std = 1.0
        else:
            input_mean = 127.5
            input_std = 127.5
        self.input_mean = input_mean
        self.input_std = input_std
        self.session = onnxruntime.InferenceSession(self.model_file, providers=providers)
        input_cfg = self.session.get_inputs()[0]
        input_shape = input_cfg.shape
        input_name = input_cfg.name
        self.input_size = tuple(input_shape[2:4][::-1])
        self.input_shape = input_shape
        outputs = self.session.get_outputs()
        output_names = []
        for out in outputs:
            output_names.append(out.name)
        self.input_name = input_name
        self.output_names = output_names
        assert len(self.output_names) == 1
        self.output_shape = outputs[0].shape

    def prepare(self, ctx_id, **kwargs):
        if ctx_id < 0:
            self.session.set_providers(['CPUExecutionProvider'])

    def get(self, img, face):
        aimg = face_align.norm_crop(img, landmark=face.kps, image_size=self.input_size[0])
        face.embedding = self.get_feat(aimg).flatten()
        return face.embedding

    def compute_sim(self, feat1, feat2):
        from numpy.linalg import norm
        feat1 = feat1.ravel()
        feat2 = feat2.ravel()
        sim = np.dot(feat1, feat2) / (norm(feat1) * norm(feat2))
        return sim

    def get_feat(self, imgs):
        if not isinstance(imgs, list):
            imgs = [imgs]
        input_size = self.input_size

        blob = cv2.dnn.blobFromImages(imgs, 1.0 / self.input_std, input_size,
                                      (self.input_mean, self.input_mean, self.input_mean), swapRB=True)
        net_out = self.session.run(self.output_names, {self.input_name: blob})[0]
        return net_out

    def forward(self, batch_data):
        blob = (batch_data - self.input_mean) / self.input_std
        net_out = self.session.run(self.output_names, {self.input_name: blob})[0]
        return net_out




# ordinary imports
from typing import List, Union
import numpy as np

import onnxruntime


from face2face.core.compatibility.Face import Face
from face2face.core.compatibility.FaceAnalysis import FaceAnalysis
from face2face.core.compatibility.INSwapper import INSwapper
from face2face.core.mixins._image_swap import _ImageSwap
from media_toolkit import media_from_file, VideoFile, ImageFile

from face2face.core.mixins._face_embedding import _FaceEmbedding
from face2face.core.mixins._face_enhance import _FaceEnhancer
from face2face.core.mixins._face_recognition import _FaceRecognition
from face2face.core.mixins._video_swap import _Video_Swap

from face2face.core.modules.utils.utils import load_image
from face2face.core.modules.utils.utils import download_model
from face2face.settings import EMBEDDINGS_DIR, DEVICE_ID


class Face2Face(_ImageSwap, _FaceEmbedding, _FaceRecognition, _Video_Swap, _FaceEnhancer):
    def __init__(self, face_embedding_folder: str = None, device_id: int = None):
        """
        :param model_path: the folder where the models are stored and downloaded to.
            results in structure like models/insightface/inswapper_128.onnx model
            and models/face_enhancer/gfpgan_1.4.onnx
        :param inswapper_model_name:
        """
        # download inswapper model (roop) if not existing and insightface models
        swapper_model_file_path = download_model("inswapper_128")
        face_analyiser_models_path = download_model("buffalo_l")

        self.providers = onnxruntime.get_available_providers()
        # Setting GPU number and creating onnx session
        if device_id is None:
            device_id = DEVICE_ID

        if "CUDAExecutionProvider" in self.providers:
            self.providers.remove("CUDAExecutionProvider")
            self.providers.append(("CUDAExecutionProvider", {'device_id': device_id}))
            self.providers = [("CUDAExecutionProvider", {'device_id': device_id})]

        self._face_analyser = FaceAnalysis(model_dir=face_analyiser_models_path, providers=self.providers)
        self._face_analyser.prepare(ctx_id=0, det_size=(320, 320))

        self._face_swapper = INSwapper(model_file=swapper_model_file_path, providers=self.providers)

        # face swapper has the option to swap images from previously stored faces as embeddings
        # they dict has structure {faces: face_embedding }
        if face_embedding_folder is None:
            face_embedding_folder = EMBEDDINGS_DIR

        self._face_embedding_folder = face_embedding_folder
        self._face_embeddings = {}


    def swap(
        self,
        media: Union[str, np.ndarray, tuple, List[str], ImageFile, VideoFile],
        faces: Union[str, dict, list, List[Face], Face, None] = None,
        enhance_face_model: Union[str, None] = 'gpen_bfr_512',
        include_audio: bool = True
    ) -> Union[np.array, list, VideoFile]:
        """
        This is a unified function to swap faces in images and videos.
        :param media: the image or video to swap the faces in.
            if str -> path to image or video. Load from file
            if np.array -> image
            if tuple -> (image, image). Same as swap_one
            if List[str] -> list of file paths. Perform swap on all of those files
            if generator -> generator that yields images or tuples (image, audio)
            if VideoFile -> video file

        :param faces: defines what to swap to.
            if str -> the name of the face to swap to. All faces in the image will be swapped to this face.
            if list -> the list of face_names or face objects to swap to from left to right.
            if dict -> the swap_pairs with the structure {source_face_name: target_face_name}. Use face recognition
                to swap the source faces to the target faces.
            if Face -> the face object to swap to. All faces in the image will be swapped to this face.

        :param enhance_face_model: the face enhancement model to use. Use None for no enhancement
        :param include_audio: if True, the audio will be included in the output video if the input is a video.
        """
        if isinstance(media, tuple):
            return self.swap_img_to_img(media[0], media[1], enhance_face_model)
        if isinstance(media, list):
            return [self.swap(inp, faces, enhance_face_model, include_audio) for inp in media]

        if faces is None:
            raise ValueError("Please provide faces to swap to.")

        # convert to image or video file
        file = None
        if type(media) in [ImageFile, VideoFile]:
            file = media
        elif isinstance(media, str):
            file = media_from_file(media)
        elif isinstance(media, np.ndarray):
            file = ImageFile().from_np_array(media)

        # perform swaps
        if isinstance(file, ImageFile):
            return self.swap_image(image=file, faces=faces, enhance_face_model=enhance_face_model)
        elif isinstance(file, VideoFile):
            return self.swap_video(
                faces=faces, video=file, include_audio=include_audio, enhance_face_model=enhance_face_model
            )

        raise ValueError(f"Wrong file type {media}. Check input.")

    def detect_faces(self, image: Union[np.array, str, ImageFile]) -> Union[List | None]:
        """
        get faces from left to right by order
        """
        image = load_image(image)
        try:
            face = self._face_analyser.get(image)
            return sorted(face, key=lambda x: x.bbox[0])
        except IndexError:
            return None


